import os
from typing import *

import requests

from parsee.templates.job_template import JobTemplate
from parsee.templates.template_from_json import from_json_dict
from parsee.extraction.extractor_elements import StandardDocumentFormat
from parsee.converters.json_to_raw import load_document_from_json
from parsee.extraction.extractor_dataclasses import ParseeAnswer, ParseeMeta, source_from_json


class ParseeCloud:

    def __init__(self, api_key: str, custom_host: Optional[str] = None):

        self.api_key = api_key
        self.host = custom_host if custom_host is not None else (os.getenv("BACKEND_HOST") if os.getenv("BACKEND_HOST") is not None else "https://backend.parsee.ai")

    def _headers(self):
        return {"Authorization": self.api_key}

    def get_template(self, template_id: str) -> JobTemplate:
        url = f"{self.host}/api/extraction/template/id/{template_id}"
        template_request = requests.get(url, headers=self._headers())
        if template_request.content == b'':
            raise Exception("template not found or not accessible")
        return from_json_dict(template_request.json())

    def get_document(self, source_identifier: str) -> StandardDocumentFormat:
        url = f"{self.host}/api/document/get-json-document?identifier={source_identifier}"
        data = requests.get(url, headers=self._headers()).json()
        return load_document_from_json(data)

    def save_template(self, template: JobTemplate, public: bool = False) -> str:
        url = f"{self.host}/api/extraction/template"
        template_json_dict = template.to_json_dict()
        template_json_dict = {**template_json_dict, "public": public}
        request = requests.post(url, json=template_json_dict, headers=self._headers())
        return request.text

    # currently only output for general questions is supported here (i.e. textual output)
    def get_output(self, source_identifier: str, template_id: str) -> List[ParseeAnswer]:
        url = f"{self.host}/api/extraction/output/json/{source_identifier}"
        data = requests.get(url, headers=self._headers()).json()
        output = []
        for entry in data:
            if entry["templateId"] == template_id and entry["text"] is not None:
                class_id = entry["detectedClass"]
                class_value = entry["text"]["answerParsed"]
                meta_answers = []
                for item in entry["text"]["metaInfo"]:
                    meta_answers.append(ParseeMeta(item["model"], 0, [], item["class_id"], item["value"], item["prob"]))
                sources = [source_from_json(x) for x in entry["sources"]]
                output.append(ParseeAnswer(entry["model"], sources, class_id, class_value, "", True, meta_answers))
        return output



from typing import Tuple, List, Union
from decimal import Decimal

from parsee.utils.enums import DocumentType
from parsee.extraction.extractor_elements import ExtractedEl


class RawToJsonConverter:

    def __init__(self, doc_type: DocumentType):
        self.doc_type = doc_type

    def convert(self, file_path_or_content: str) -> Tuple[List[ExtractedEl], Decimal]:
        raise NotImplemented



from pydantic.dataclasses import dataclass
from typing import Dict, Optional, List, Union

from parsee.templates.element_schema import ElementDetectionSchema
from parsee.templates.general_structuring_schema import GeneralQuerySchema, StructuringItemSchema
from parsee.extraction.models.model_dataclasses import MlModelSpecification


@dataclass
class JobTemplate:
    id: Union[None, str]
    title: str
    description: str
    questions: GeneralQuerySchema
    detection: ElementDetectionSchema
    meta: List[StructuringItemSchema]
    general_settings: Optional[Dict] = None

    # sets all models to one default model
    def set_default_model(self, model: MlModelSpecification):
        for item in self.questions.items:
            item.model = model.internal_name
        for item in self.detection.items:
            item.model = model.internal_name
            if item.mapRows is not None:
                item.mappingModel = model.internal_name
        for item in self.meta:
            item.model = model.internal_name

    def to_json_dict(self) -> Dict:
        return {"id": self.id, "title": self.title, "description": self.description, "questions": self.questions.to_json_dict(), "detection": self.detection.to_json_dict(), "meta": [x.to_json_dict() for x in self.meta]}


from typing import *
from functools import reduce

from parsee.templates.job_template import JobTemplate
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.storage.interfaces import StorageManager
from parsee.extraction.extractor_elements import StandardDocumentFormat, FinalOutputTableColumn
from parsee.extraction.extractor_dataclasses import ParseeAnswer, ParseeBucket
from parsee.storage.in_memory_storage import InMemoryStorageManager
from parsee.extraction.final_structuring import get_structured_tables_from_locations, final_tables_from_columns
from parsee.extraction.models.model_loader import element_models_from_schema, meta_models_from_items, question_models_from_schema, mapping_models_from_schema, ModelLoader


def run_job_with_single_model(doc: StandardDocumentFormat, job_template: JobTemplate, model: MlModelSpecification, custom_model_loader: Optional[ModelLoader] = None) -> Tuple[List[ParseeBucket], List[FinalOutputTableColumn], List[ParseeAnswer]]:
    model_loader = custom_model_loader
    if model_loader is None:
        storage = InMemoryStorageManager([model])
        model_loader = ModelLoader(storage)
    # update the models
    job_template.set_default_model(model)
    return structure_data(doc, job_template, model_loader, {})


def structure_data(doc: StandardDocumentFormat, job_template: JobTemplate, model_loader: ModelLoader, params: Dict[str, Any]) -> Tuple[List[ParseeBucket], List[FinalOutputTableColumn], List[ParseeAnswer]]:

    # add manual answers to params
    params = {**params, **job_template.detection.settings, **job_template.questions.settings}

    question_models = question_models_from_schema(job_template.questions, job_template.meta, model_loader, params)

    answers: List[ParseeAnswer] = []
    for question_model in question_models:
        answers += question_model.predict_answers(doc)

    models_loc = element_models_from_schema(job_template.detection, model_loader, params)

    locations = []
    for model_loc in models_loc:
        locations += model_loc.classify_elements(doc)

    output_values = get_structured_tables_from_locations(job_template, doc, locations)

    # add meta values
    all_meta_ids = list(set(reduce(lambda acc, x: acc+x.metaInfoIds, job_template.detection.items, [])))
    meta_ids_by_main_class = reduce(lambda acc, x: {**acc, x.id: x.metaInfoIds}, job_template.detection.items, {})
    if len(all_meta_ids) > 0:
        meta_models = meta_models_from_items([x for x in job_template.meta if x.id in all_meta_ids], model_loader, params)
        for meta_model in meta_models:
            meta_predictions_list = meta_model.predict_meta(output_values, doc.elements)
            for k, meta_predictions in enumerate(meta_predictions_list):
                output_values[k].meta += [x for x in meta_predictions if x.class_id in meta_ids_by_main_class[output_values[k].detected_class]]

    # run mapping
    all_mappings: List[ParseeBucket] = []
    tables = final_tables_from_columns(output_values)
    mapping_models = mapping_models_from_schema(job_template.detection, model_loader, params)
    for model_mapping in mapping_models:
        for table in tables:
            mappings, mapping_schema = model_mapping.classify_elements(table)
            if mapping_schema is not None:
                for col in table.columns:
                    col.apply_mappings(mappings, mapping_schema)
            all_mappings += mappings

    return all_mappings, output_values, answers


from typing import *
import re

from parsee.extraction.tasks.element_classification.element_model import ElementModel, ElementSchema, StandardDocumentFormat, ParseeLocation
from parsee.storage.interfaces import StorageManager
from parsee.utils.helper import is_number_cell, clean_numeric_value, parse_json_array
from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel
from parsee.extraction.tasks.element_classification.features import LLMLocationFeatureBuilder


class ElementModelLLM(ElementModel):

    def __init__(self, items: List[ElementSchema], storage: StorageManager, llm: LLMBaseModel, **kwargs):
        super().__init__(items)
        self.model_name = llm.model_name
        self.llm = llm
        self.max_search_items = 10
        self.storage = storage
        self.prob = 0.8
        self.feature_builder: LLMLocationFeatureBuilder = LLMLocationFeatureBuilder()

    def parse_prompt_answer(self, prompt_answer: str) -> List[int]:
        answers = prompt_answer.splitlines()
        output = []
        for k, answer in enumerate(answers):
            values_predicted = parse_json_array(answer)
            if values_predicted is not None:
                for val in values_predicted:
                    if isinstance(val, int):
                        output.append(val)
        return output

    def classify_elements(self, document: StandardDocumentFormat) -> List[ParseeLocation]:

        output: List[ParseeLocation] = []

        for item in self.items:

            prompt = self.feature_builder.make_prompt(item, document, self.storage)

            answer, amount = self.llm.make_prompt_request(str(prompt))
            self.storage.log_expense(self.llm.model_name, amount, item.id)

            best_indexes = self.parse_prompt_answer(answer)

            partial_prob = 0.0 if len(best_indexes) <= 1 else 0.6

            for best_idx in best_indexes:
                if 0 <= best_idx < len(document.elements):
                    el = document.elements[best_idx]
                    location = ParseeLocation(self.model_name, partial_prob, item.id, self.prob, el.source, [])
                    output.append(location)
        return output



import shutil
from typing import *
import os
import csv
import uuid
import pickle
from shutil import rmtree

from parsee.datasets.dataset_dataclasses import BaseDatasetRow, Transformation, ColumnSettings
from parsee.datasets.writers.interfaces import DatasetWriter, ModelWriter


class CsvDiskWriter(DatasetWriter):

    def __init__(self, write_location: str):

        # create folder
        folder_name = "dataset_"+str(uuid.uuid4())
        final_path = os.path.join(write_location, folder_name)
        os.mkdir(final_path)
        self.base_location = write_location
        self.write_location = final_path
        self.file = {}
        self.writer = {}
        self.rows_written = {}
        self.full_paths = {}

    def file_name(self, dataset_name: str):
        return dataset_name+".csv"

    def create_dataset_file(self, dataset: str):

        full_file_path = os.path.join(self.write_location, self.file_name(dataset))
        self.full_paths[dataset] = full_file_path
        self.file[dataset] = open(full_file_path, "w", newline='', encoding="utf8")
        self.writer[dataset] = csv.writer(self.file[dataset], delimiter=',', quotechar='"')
        self.rows_written[dataset] = 0

    def _write_rows(self, dataset_rows: List[BaseDatasetRow], dataset: str):

        if len(dataset_rows) == 0:
            return

        if dataset not in self.file:
            self.create_dataset_file(dataset)

        if self.rows_written[dataset] == 0:
            self.writer[dataset].writerow(dataset_rows[0].column_names())

        self.writer[dataset].writerows([x.to_list() for k, x in enumerate(dataset_rows)])
        self.rows_written[dataset] += len(dataset_rows)

    def finish_writing(self):
        # close files
        for _, file in self.file.items():
            file.close()

    def delete_all_files(self):
        shutil.rmtree(self.write_location)




from typing import *

from parsee.extraction.extractor_elements import FinalOutputTableColumn, ExtractedEl, StructuredTable
from parsee.templates.general_structuring_schema import StructuringItemSchema
from parsee.extraction.extractor_dataclasses import ParseeMeta
from parsee.utils.helper import is_number_cell, is_year_cell, clean_text_for_word_vectors2
from parsee.datasets.dataset_dataclasses import DatasetRow, MetaUniqueIdentifier
from parsee.extraction.models.llm_models.prompts import Prompt
from parsee.extraction.models.llm_models.structuring_schema import get_prompt_schema_item


class MetaFeatureBuilder:

    memory: Dict[MetaUniqueIdentifier, DatasetRow]

    def __init__(self):
        self.memory = {}
        self.cell_start_delimiter = "xcellstartx"
        self.row_start_delimiter = "xrowstartx"
        self.cell_self_append = "xselfx"
        self.number_replacement = "xnumberx"

    # returns features for meta info ml of table using word embeddings
    def get_meta_features_table(self, structured_table: StructuredTable, col_index_chosen: int, text_before: int, base_year: Union[int, None]) -> Dict[str, any]:

        def get_words_from_cells(cell_obj_list, repeat=False):
            objects_used = []
            all_fragments = []
            for c_obj in cell_obj_list:
                if type(c_obj) is str:
                    all_fragments.append(c_obj)
                else:
                    if (c_obj not in objects_used or repeat) and c_obj.val is not None:
                        all_fragments.append(c_obj.val)
                        objects_used.append(c_obj)

            all_text = " ".join(all_fragments)
            return all_text

        # collect text from various places
        words_list = {"before": text_before, "li_and_values": "", "top_li": "", "top_data": "", "top_column_all": "", "top_column_pos_0": "", "top_column_pos_1": "", "top_column_pos_2": "",
                      "other_li": "", "other_values": "", "neighbor_col_top_left": "", "neighbor_col_top_right": "", "total_header_values": "", "top_li_reworked": ""}

        # determine words of line items and values
        cell_list = []
        for row_index in structured_table.numeric_rows_indices:
            for cell_obj in structured_table.rows[row_index].final_values:
                # exclude numbers
                if is_number_cell(cell_obj.val) and not is_year_cell(cell_obj.val):
                    continue
                cell_list.append(cell_obj)
        words_list['li_and_values'] = get_words_from_cells(cell_list)

        # determine words on top of line items
        cell_list = []
        for row in structured_table.header_rows:
            for col_index, cell_obj in enumerate(row.final_values):
                if col_index < min(structured_table.numeric_cols_indices):
                    cell_list.append(cell_obj)
        words_list['top_li'] = get_words_from_cells(cell_list)

        # determine words on top of data
        cell_list = []
        for row in structured_table.header_rows:
            for col_index, cell_obj in enumerate(row.final_values):
                if col_index >= min(structured_table.numeric_cols_indices):
                    cell_list.append(cell_obj)
        words_list['top_data'] = get_words_from_cells(cell_list)

        # top line items reworked
        cell_list = []
        for row in structured_table.header_rows:
            if row.has_values(max_col_index=min(structured_table.numeric_cols_indices) - 1):
                for col_index, cell_obj in enumerate(row.final_values):
                    if col_index < min(structured_table.numeric_cols_indices):
                        # check that cell obj is not contained in first value column
                        if cell_obj != row.final_values[structured_table.numeric_cols_indices[0]]:
                            cell_list.append(cell_obj)
        words_list['top_li_reworked'] = get_words_from_cells(cell_list)

        # total header data
        cell_list = []
        for row in structured_table.header_rows:
            if row.has_values():
                cell_list.append(self.row_start_delimiter)
                for col_index, cell_obj in enumerate(row.final_values):
                    if col_index in structured_table.numeric_cols_indices:
                        delim = self.cell_start_delimiter if col_index != col_index_chosen else self.cell_start_delimiter + self.cell_self_append
                        cell_list.append(delim)
                        cell_list.append(cell_obj)
        words_list['total_header_values'] = get_words_from_cells(cell_list, True)

        # determine words on top IN COLUMN
        cell_list = []
        for row in structured_table.header_rows:
            cell_obj = row.final_values[col_index_chosen]
            cell_list.append(cell_obj)
        words_list['top_column_all'] = get_words_from_cells(cell_list)

        # determine words on top IN COLUMN, pos 0 only
        cell_list = []
        start_index = min(structured_table.numeric_rows_indices) - 1 if len(structured_table.numeric_rows_indices) > 0 else 0
        to_skip = 0
        to_skip_counter = -1
        for row_index in range(start_index, -1, -1):
            cell = structured_table.rows[row_index].final_values[col_index_chosen]
            if cell.val is not None and cell.val.strip() != "":
                to_skip_counter += 1
                if to_skip <= to_skip_counter:
                    cell_list.append(cell)
                    break
        words_list['top_column_pos_0'] = get_words_from_cells(cell_list)

        # determine words on top IN COLUMN, pos 1 only
        cell_list = []
        start_index = min(structured_table.numeric_rows_indices) - 1 if len(structured_table.numeric_rows_indices) > 0 else 0
        to_skip = 1
        to_skip_counter = -1
        for row_index in range(start_index, -1, -1):
            cell = structured_table.rows[row_index].final_values[col_index_chosen]
            if cell.val is not None and cell.val.strip() != "":
                to_skip_counter += 1
                if to_skip <= to_skip_counter:
                    cell_list.append(cell)
                    break
        words_list['top_column_pos_1'] = get_words_from_cells(cell_list)

        # determine words on top IN COLUMN, pos 2 only
        cell_list = []
        start_index = min(structured_table.numeric_rows_indices) - 1 if len(structured_table.numeric_rows_indices) > 0 else 0
        to_skip = 2
        to_skip_counter = -1
        for row_index in range(start_index, -1, -1):
            cell = structured_table.rows[row_index].final_values[col_index_chosen]
            if cell.val is not None and cell.val.strip() != "":
                to_skip_counter += 1
                if to_skip <= to_skip_counter:
                    cell_list.append(cell)
                    break
        words_list['top_column_pos_2'] = get_words_from_cells(cell_list)

        # other locations, li
        cell_list = []
        for row in structured_table.other_rows:
            for col_index, cell_obj in enumerate(row.final_values):
                if col_index < min(structured_table.numeric_cols_indices):
                    cell_list.append(cell_obj)
        words_list['other_li'] = get_words_from_cells(cell_list)

        # other locations, values
        cell_list = []
        for row in structured_table.other_rows:
            for col_index, cell_obj in enumerate(row.final_values):
                if col_index >= min(structured_table.numeric_cols_indices):
                    cell_list.append(cell_obj)
        words_list['other_values'] = get_words_from_cells(cell_list)

        # neighbor col top left
        col_indices_available = list(sorted(structured_table.numeric_cols_indices))
        idx_self = col_indices_available.index(col_index_chosen)
        if idx_self > 0:
            cell_list = []
            for row in structured_table.header_rows:
                cell_obj = row.final_values[col_indices_available[idx_self - 1]]
                cell_list.append(cell_obj)
            words_list['neighbor_col_top_left'] = get_words_from_cells(cell_list)

        # neighbor col top right
        col_indices_available = list(sorted(structured_table.numeric_cols_indices))
        idx_self = col_indices_available.index(col_index_chosen)
        if idx_self < len(col_indices_available) - 1:
            cell_list = []
            for row in structured_table.header_rows:
                cell_obj = row.final_values[col_indices_available[idx_self + 1]]
                cell_list.append(cell_obj)
            words_list['neighbor_col_top_right'] = get_words_from_cells(cell_list)

        # make final values
        for key, values in words_list.items():
            words_list[key] = clean_text_for_word_vectors2(values, base_year, number_token=self.number_replacement)

        return words_list
    
    def make_features(self, source_identifier: Optional[str], template_id: Optional[str], column: FinalOutputTableColumn, elements: List[ExtractedEl], base_year: Optional[int], custom_features: Optional[Dict[str, any]]) -> DatasetRow:

        unique_identifier = MetaUniqueIdentifier(column.detected_class, column.col_idx, column.kv_identifier)

        if unique_identifier in self.memory:
            return self.memory[unique_identifier]

        text_words_to_include_before = 200
        main_el: StructuredTable = elements[column.locations[0].source.element_index]

        text_before_words = main_el.get_text_before(elements, text_words_to_include_before)
        org_idx = column.org_col_indices[0]
        features = self.get_meta_features_table(main_el, org_idx, text_before_words, base_year)

        features = {"col_idx_values": column.col_idx, "col_idx_all": org_idx, **features}

        if custom_features is not None:
            features = {**features, **custom_features}

        self.memory[unique_identifier] = DatasetRow(source_identifier, template_id, unique_identifier, features)

        return self.memory[unique_identifier]


class LLMMetaFeatureBuilder(MetaFeatureBuilder):

    def __init__(self):
        super().__init__()
        self.cell_start_delimiter = "(cell start)"
        self.row_start_delimiter = "(row start)"
        self.cell_self_append = "(cell main)"
        self.number_replacement = "[number]"

    def build_raw_answer(self, items: List[StructuringItemSchema], results: List[ParseeMeta]) -> str:
        output = ""
        for k, item in enumerate(items):
            if k > 0:
                output += "\n"
            schema_item = get_prompt_schema_item(item)
            meta_items_filtered = [x for x in results if x.class_id == item.id]
            val = "n/a" if len(meta_items_filtered) == 0 else schema_item.parsed_to_raw(meta_items_filtered[0].class_value)
            output += f"{k+1}) {val}"
        return output

    def text_main(self, items: List[StructuringItemSchema]) -> str:
        output = ""

        for k, item in enumerate(items):
            schema_item = get_prompt_schema_item(item)
            additional_info_str = f" Additional info: {item.additionalInfo}." if item.additionalInfo.strip() != "" else ""
            output += f"{k + 1}) {item.title}.{additional_info_str} {schema_item.get_possible_values_str()}\n"

        return output

    def example(self, items: List[StructuringItemSchema]) -> str:
        output = "Your answer could be for example: \n"
        for k, item in enumerate(items):
            output += f"{k + 1}) {get_prompt_schema_item(item).get_example()}\n"

        return output

    def build_prompt(self, features: DatasetRow, col_idx: int, items: List[StructuringItemSchema]) -> Prompt:

        available_data = f'''
                The current column index is: {col_idx}. \n
                On top of the column (in the header part), there is this text: "{features.get_feature('top_column_all')}"\n
                The content of the table is the following (excluding any numbers): {features.get_feature('li_and_values')}\n
                On top of the line items, there is this text: "{features.get_feature('top_li_reworked')}\n"

                Before the table starts, there is this text: "{features.get_feature('before')}"\n
                '''

        prompt = Prompt("We want to recognize certain information in a table, more specifically in a single column.", f'Please identify the following information based on the provided text: {self.text_main(items)}',
                        'It is very important that you only answer with the number of the question and then one of the valid values for that question afterwards.', self.example(items), available_data)

        return prompt

    def make_prompt(self, column: FinalOutputTableColumn, elements: List[ExtractedEl], items: List[StructuringItemSchema]) -> Prompt:

        features = self.make_features(None, None, column, elements, None, None)

        return self.build_prompt(features, column.col_idx, items)


import os
import time
from typing import List, Tuple
from dataclasses import dataclass
from decimal import Decimal
import math

import anthropic
import tiktoken

from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel, truncate_prompt
from parsee.extraction.models.model_dataclasses import MlModelSpecification


class AnthropicModel(LLMBaseModel):

    def __init__(self, model: MlModelSpecification):
        super().__init__(model.name)
        self.model = model
        self.max_retries = 5
        self.encoding = tiktoken.get_encoding("cl100k_base")
        self.max_tokens_answer = 1024
        self.max_tokens_question = self.model.max_tokens - self.max_tokens_answer
        self.client = anthropic.Anthropic(api_key=model.api_key if model.api_key is not None else os.getenv("ANTHROPIC_API_KEY"))

    def _call_api(self, prompt: str, retries: int = 0, wait: int = 5) -> Tuple[str, Decimal]:
        try:
            message = self.client.messages.create(
                model=self.model.internal_name,
                max_tokens=self.max_tokens_answer,
                temperature=0,
                messages=[
                    {"role": "user", "content": prompt}
                ]
            )

            answer = message.content[0].text if len(message.content) > 0 else ""
            final_cost = (Decimal(message.usage.input_tokens+message.usage.output_tokens) * Decimal(self.model.price_per_1k_tokens/1000)) if self.model.price_per_1k_tokens is not None else Decimal(0)
            return answer, final_cost
        except Exception as e:
            if retries < self.max_retries:
                time.sleep(wait * 2 ** retries)
                return self._call_api(prompt, retries + 1)
            else:
                return "", Decimal(0)

    def make_prompt_request(self, prompt: str) -> Tuple[str, Decimal]:
        final_prompt, _ = truncate_prompt(prompt, self.encoding, self.max_tokens_question)
        return self._call_api(final_prompt)



from typing import *
from functools import reduce

from parsee.templates.job_template import JobTemplate
from parsee.templates.general_structuring_schema import GeneralQueryItemSchema, GeneralQuerySchema, StructuringItemSchema
from parsee.templates.element_schema import ElementDetectionSchema, ElementSchema
from parsee.utils.enums import OutputType, ContextType, SearchStrategy


id_store = {}


# creates some simple IDs for meta items etc (shorter IDs work in general better for LLM models than some long hashes for example)
def simple_id(category: str):
    if category not in id_store:
        id_store[category] = 0
    id_store[category] += 1
    return f"{category}{id_store[category]}"


class MetaItem(StructuringItemSchema):

    def __init__(self, question: str, output_type: OutputType, list_values: Optional[List[str]] = None, assigned_id: Optional[str] = None, example: Optional[str] = None, additional_info: Optional[str] = None):

        self.model = None
        self.type = output_type
        self.context = ContextType.QUESTIONS
        self.id = simple_id("meta") if assigned_id is None else assigned_id
        self.title = question
        self.additionalInfo = "" if additional_info is None else additional_info
        self.searchStrategy = SearchStrategy.VECTOR
        self.valuesList = list_values
        self.example = example
        self.keywords = None
        self.defaultValue = None
        self.customArgsJson = None


class StructuringItem(GeneralQueryItemSchema):

    def __init__(self, question: str, output_type: OutputType, list_values: Optional[List[str]] = None, meta_info: Optional[List[MetaItem]] = None, assigned_id: Optional[str] = None, example: Optional[str] = None, additional_info: Optional[str] = None):

        self.model = None
        self.type = output_type
        self.context = ContextType.QUESTIONS
        self.id = simple_id("question") if assigned_id is None else assigned_id
        self.title = question
        self.additionalInfo = "" if additional_info is None else additional_info
        self.searchStrategy = SearchStrategy.VECTOR
        self.valuesList = list_values
        self.metaInfoIds = [] if meta_info is None else [x.id for x in meta_info]
        self.meta_info = meta_info
        self.example = example
        self.keywords = None
        self.valuesList = None
        self.defaultValue = None
        self.customArgsJson = None


class RowMapping:

    def __init__(self):
        pass # TODO


class TableItem(ElementSchema):

    def __init__(self, item_title: str, keywords: str, meta_info: Optional[List[MetaItem]] = None, row_mapping: Optional[RowMapping] = None, item_description: str = "", assigned_id: Optional[str] = None):

        self.id = simple_id("table") if assigned_id is None else assigned_id
        self.title = item_title
        self.additionalInfo = item_description
        self.keywords = keywords
        self.takeBestInProximity = False
        self.model = None
        self.searchStrategy = SearchStrategy.VECTOR
        self.mapRows = row_mapping
        self.mappingModel = None
        self.metaInfoIds = [] if meta_info is None else [x.id for x in meta_info]
        self.meta_info = meta_info


def create_template(structuring_items: Optional[List[StructuringItem]], table_items: Optional[List[TableItem]] = None) -> JobTemplate:

    structuring_items = [] if structuring_items is None else structuring_items
    table_items = [] if table_items is None else table_items
    meta_items_general = reduce(lambda acc, x: acc+x, [x.meta_info for x in structuring_items if x.meta_info is not None], [])
    meta_items_tables = reduce(lambda acc, x: acc + x, [x.meta_info for x in table_items if x.meta_info is not None], [])
    all_meta_items = list(set(meta_items_tables + meta_items_general))

    return JobTemplate(None, simple_id("template"), "", GeneralQuerySchema(structuring_items, {}), ElementDetectionSchema(table_items, {}), all_meta_items, None)



import json
from typing import Dict
from pydantic.tools import parse_obj_as

from parsee.templates.job_template import JobTemplate


def from_json(json_string: str) -> JobTemplate:

    data = json.loads(json_string)

    job = parse_obj_as(JobTemplate, data)

    return job


def from_json_dict(data: Dict) -> JobTemplate:

    job = parse_obj_as(JobTemplate, data)

    return job
